@@ -17,46 +17,80 @@ class InstaClient:
     CHROMEDRIVER=1
     LOCAHOST=1
     WEB_SERVER=2
-    def __init__(self, driver_type: int=CHROMEDRIVER, host=None):
+    def __init__(self, driver_type: int=CHROMEDRIVER, host:int=LOCAHOST):
         """
-        Creates an instance of instaclient class.
+        Create an `InstaClient` object to access the instagram website.
 
         Args:
-            driver_type:int: 
-                InstaClient.CHROMEDRIVER for a Chrome instance
-            host:int: 
-                InstaClient.LOCALHOST for testing / running locally
-                InstaClient.WEB_SERVER for production / running on a web server
+            driver_type (int, optional): The type of browser driver to run instagram on. Defaults to CHROMEDRIVER.
+            host (int, optional): Whether the code is run locally or on a server. Defaults to LOCAHOST.
 
-        Attributes:
-            driver: Instance of the Selenium Webdriver (defaults to Chrome) 
-            logged_in:bool: Boolean whether current user is logged in or not. Defaults to False
+        Raises:
+            InvaildHostError: Raised if host int does not correspond to any host type
+            InvaildDriverError: Raised if driver int does not correspond to any driver type.
+            error: Normal Exception, raised if anything fails when creating the client.
         """
         try:
             if driver_type == self.CHROMEDRIVER:
-                if host in (None, self.LOCAHOST):
+                if host == self.WEB_SERVER:
                     # Running on web server
                     chrome_options = webdriver.ChromeOptions()
                     chrome_options.binary_location = os.environ.get("GOOGLE_CHROME_BIN")
+                    chrome_options.add_argument('--user-agent=Mozilla/5.0 (iPhone; CPU iPhone OS 10_3 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) CriOS/56.0.2924.75 Mobile/14E5239e Safari/602.1')
+                    chrome_options.add_argument("window-size=525,950")
                     chrome_options.add_argument("--headless")
                     chrome_options.add_argument("--disable-dev-shm-usage")
                     chrome_options.add_argument("--no-sandbox")
                     self.driver = webdriver.Chrome(executable_path=os.environ.get("CHROMEDRIVER_PATH"), chrome_options=chrome_options)
-                else:
+                elif host == self.LOCAHOST:
                     # Running locally
-                    self.driver = webdriver.Chrome('instaclient/drivers/chromedriver.exe')
+                    chrome_options = webdriver.ChromeOptions()
+                    chrome_options.add_argument('--user-agent=Mozilla/5.0 (iPhone; CPU iPhone OS 10_3 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) CriOS/56.0.2924.75 Mobile/14E5239e Safari/602.1')
+                    chrome_options.add_argument("window-size=525,950")
+                    self.driver = webdriver.Chrome(executable_path='instaclient/drivers/chromedriver.exe', chrome_options=chrome_options)
+                else:
+                    raise InvaildHostError(host)
             else:
                 raise InvaildDriverError(driver_type)
-            self.driver.maximize_window()
         except Exception as error:
             raise error
         self.logged_in = False
 
 
     @insta_method
-    def login(self, username:str, password:str):
+    def check_status(self):
         """
-        Logs a user into Instagram via the web portal
+        Check if account is currently logged in. Returns True if account is logged in. Sets the instaclient.logged_in variable accordingly.
+
+        :return: True if account is logged in, False if account is NOT logged in.
+        :rtype: boolean
+        """
+        icon = self.__check_existence(EC.presence_of_element_located((By.XPATH, Paths.NAV_BAR)), wait_time=4)
+        if icon:
+            self.logged_in = True
+            return True
+        else:
+            self.logged_in = False
+            return False
+
+
+    @insta_method
+    def login(self, username:str, password:str, check_user:bool=True):
+        """
+        Sign Into Instagram with credentials. Go through 2FA if necessary. Sets the InstaClient variable `InstaClient.logged_in` to True if login was successful.
+
+        Args:
+            username (str): The instagram account's username
+            password (str): The instagram account's password
+            check_user (bool, optional): If False, the username will be considered as valid and will not be checked. If the user is invalid, the login procedure will not be completed. Defaults to True.
+
+        Raises:
+            InvalidUserError: Raised if the user is not valid and `check_user` is set to True. Warning: if check_user is set to False and the user is invalid, the login procedure will not be completed.
+            InvaildPasswordError: Raised if the password is incorrect.
+            SecurityCodeNecessary: Raised if the user's account has 2FA. If this error is raised, you can complete the login procedure with `InstaClient.input_security_code`
+
+        Returns:
+            bool: Returns True if login was successful.
         """
         self.username = username
         self.password = password
@@ -66,15 +100,15 @@ class InstaClient:
             self.driver.get(ClientUrls.LOGIN_URL)
             # Detect Cookies Dialogue
             try:
-                alert = self._find_element(EC.element_to_be_clickable((By.XPATH, Paths.ACCEPT_COOKIES)), wait_time=3)
+                alert = self.__find_element(EC.element_to_be_clickable((By.XPATH, Paths.ACCEPT_COOKIES)), wait_time=3)
                 alert.click()
             except:
                 print('No alert')
                 pass
             # Get Form elements
-            username_input = self._find_element(EC.presence_of_element_located((By.XPATH,Paths.USERNAME_INPUT)))
-            password_input = self._find_element(EC.presence_of_element_located((By.XPATH,Paths.PASSWORD_INPUT)))
-            login_btn = self._find_element(EC.presence_of_element_located((By.XPATH,Paths.LOGIN_BTN)))# login button xpath changes after text is entered, find first
+            username_input = self.__find_element(EC.presence_of_element_located((By.XPATH,Paths.USERNAME_INPUT)))
+            password_input = self.__find_element(EC.presence_of_element_located((By.XPATH,Paths.PASSWORD_INPUT)))
+            login_btn = self.__find_element(EC.presence_of_element_located((By.XPATH,Paths.LOGIN_BTN)))# login button xpath changes after text is entered, find first
             # Fill out form
             username_input.send_keys(username)
             time.sleep(1)
@@ -85,21 +119,25 @@ class InstaClient:
             # User already logged in ?
             print('User already logged in?')
             return self.logged_in
+        
         # Detect correct Login
-        try:
-            # Credentials Incorrect
-            alert: WebElement = self._find_element(EC.presence_of_element_located((By.XPATH,Paths.ALERT)), wait_time=3)
-            if 'username' in alert.text: #TODO insert in translation
+        if check_user:
+            usernamealert: WebElement = self.__check_existence(EC.presence_of_element_located((By.XPATH, Paths.INCORRECT_USERNAME_ALERT)), wait_time=3)
+            if usernamealert:
+                # Username is invalid
                 self.driver.get(ClientUrls.LOGIN_URL)
-                raise InvalidUserError(self.username)
-            elif 'password' in alert.text: #TODO insert in translation
-                self.driver.get(ClientUrls.LOGIN_URL)
-                raise InvaildPasswordError(self.password)
-        except (TimeoutException, NoSuchElementException):
-            pass
+                self.username = None
+                raise InvalidUserError(username)
+
+        passwordalert: WebElement = self.__check_existence(EC.presence_of_element_located((By.XPATH,Paths.INCORRECT_PASSWORD_ALERT)), wait_time=3)
+        if passwordalert:
+            # Password is incorrect
+            self.driver.get(ClientUrls.LOGIN_URL)
+            self.password = None
+            raise InvaildPasswordError(password)
 
         # Detect 2FS
-        scode_input = self._check_existence(EC.presence_of_element_located((By.XPATH, Paths.SECURITY_CODE)), wait_time=4)
+        scode_input = self.__check_existence(EC.presence_of_element_located((By.XPATH, Paths.SECURITY_CODE)), wait_time=3)
         if scode_input:
             # 2F Auth is enabled, request security code
             raise SecurityCodeNecessary()
@@ -111,7 +149,7 @@ class InstaClient:
 
         # Detect 'Turn On Notifications' Box
         try:
-            no_notifications_btn = self._find_element(EC.presence_of_element_located((By.XPATH, Paths.NO_NOTIFICATIONS_BTN)), wait_time=4)
+            no_notifications_btn = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.NO_NOTIFICATIONS_BTN)), wait_time=3)
             no_notifications_btn.click()
         except:
             pass
@@ -120,26 +158,26 @@ class InstaClient:
 
     
     @insta_method
-    def input_security_code(self, code):
+    def input_security_code(self, code:int or str):
         """
-        Use when authenticating and the 2FA security code is required.
+        Complete login procedure started with `InstaClient.login()` and insert 2FA security code. Sets `instaclient.logged_in` to True if login was successful.
 
         Args:
-            code:int: Security code sent by instagram to the user's phone number or Authenticator App.
-
-        Returns:
-            True if login is successful
+            code (int|str): The 2FA security code generated by the Authenticator App or sent via SMS to the user.
 
         Raises:
-            InvalidSecurityCodeError() if the security code is incorrect
+            InvalidSecurityCodeError: Raised if the security code is not correct
+
+        Returns:
+            bool: Returns True if login was successful
         """
-        scode_input: WebElement = self._find_element(EC.presence_of_element_located((By.XPATH, Paths.SECURITY_CODE)), wait_time=4)
+        scode_input: WebElement = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.SECURITY_CODE)), wait_time=4)
         scode_input.send_keys(code)
-        scode_btn: WebElement = self._find_element(EC.element_to_be_clickable((By.XPATH, Paths.SECURITY_CODE_BTN)), wait_time=5)
+        scode_btn: WebElement = self.__find_element(EC.element_to_be_clickable((By.XPATH, Paths.SECURITY_CODE_BTN)), wait_time=5)
         time.sleep(1)
         scode_btn.click()
 
-        alert = self._check_existence(EC.presence_of_element_located((By.XPATH, Paths.ALERT)))
+        alert = self.__check_existence(EC.presence_of_element_located((By.XPATH, Paths.ALERT)))
         if alert:
             # Code is Wrong
             # Clear input field
@@ -163,7 +201,7 @@ class InstaClient:
 
         self.nav_user(user)
 
-        follow_buttons = self._find_buttons('Follow')
+        follow_buttons = self.__find_buttons('Follow')
 
         for btn in follow_buttons:
             btn.click()
@@ -180,12 +218,12 @@ class InstaClient:
 
         self.nav_user(user)
 
-        unfollow_btns = self._find_buttons('Following')
+        unfollow_btns = self.__find_buttons('Following')
 
         if unfollow_btns:
             for btn in unfollow_btns:
                 btn.click()
-                unfollow_confirmation = self._find_buttons('Unfollow')[0]
+                unfollow_confirmation = self.__find_buttons('Unfollow')[0]
                 unfollow_confirmation.click()
         else:
             print('No {} buttons were found.'.format('Following'))
@@ -212,7 +250,7 @@ class InstaClient:
 
             finished = self._infinite_scroll() # scroll down
             
-            elements = self._find_element((EC.presence_of_element_located(By.CLASS_NAME, 'FFVAD')))
+            elements = self.__find_element((EC.presence_of_element_located(By.CLASS_NAME, 'FFVAD')))
             img_srcs.extend([img.get_attribute('src') for img in elements]) # scrape srcs
 
         img_srcs = list(set(img_srcs)) # clean up duplicates
@@ -238,7 +276,7 @@ class InstaClient:
         self.nav_user(user)
 
         imgs = []
-        elements = self._find_element(EC.presence_of_all_elements_located((By.CLASS_NAME, '_9AhH0')))
+        elements = self.__find_element(EC.presence_of_all_elements_located((By.CLASS_NAME, '_9AhH0')))
         imgs.extend(elements)
 
         for img in imgs[:n_posts]:
@@ -253,7 +291,7 @@ class InstaClient:
 
 
     @insta_method
-    def send_dm(self, user:str, message:str):
+    def send_dm(self, user:str, message:str, check_user=True):
         """
         Send a DM to the specified user.
 
@@ -262,11 +300,11 @@ class InstaClient:
             message:str: Message to send to the user
         """
         # Navigate to User's dm page
-        self.nav_user_dm(user)
-        text_area = self._find_element(EC.presence_of_element_located((By.XPATH, Paths.DM_TEXT_AREA)))
+        self.nav_user_dm(user, check_user=check_user)
+        text_area = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.DM_TEXT_AREA)))
         print(text_area)
         text_area.send_keys(message)
-        send_btn = self._find_element(EC.presence_of_element_located((By.XPATH, Paths.SEND_DM_BTN)))
+        send_btn = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.SEND_DM_BTN)))
         send_btn.click()
 
 
@@ -285,16 +323,14 @@ class InstaClient:
 
 
     @insta_method
-    def scrape_followers(self, user:str, count:int=100, callback_frequency:int=10, callback=None, *args, **kwargs):
+    def scrape_followers(self, user:str, check_user=True, *args, **kwargs):
         """
         Gets all followers of a certain user
 
         Args:
             user:str: Username of the user for followers look-up
             count:int: Number of followers to get. Note that high follower counts will take longer and longer exponentially (even hours)
-            callback_frequency:int: Number of followers to get before sending an update
-    
-        
+            
         Returns:
             followers:list<str>: List of usernames (str)
 
@@ -303,50 +339,45 @@ class InstaClient:
             PrivateAccountError if account is private
         """
         # Nav to user page
-        self.nav_user(user)
+        self.nav_user(user, check_user=check_user)
         # Find Followers button/link
-        followers_btn:WebElement = self._find_element(EC.presence_of_element_located((By.XPATH, Paths.FOLLOWERS_BTN)), wait_time=4)
-        # Get the number of followers and set the count
-        follower_count_div:WebElement = followers_btn.find_element_by_class_name(Paths.FOLLOWER_COUNT)
-        follower_count = int(follower_count_div.get_attribute('title'))
-        if count == -1:
-            # Scrape all followers:
-            count = follower_count
-        elif count > follower_count:
-            count = follower_count
-            
+        followers_btn:WebElement = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.FOLLOWERS_BTN)), wait_time=4)
         # Start scraping
         followers = []
         # Click followers btn
         followers_btn.click()
         time.sleep(2)
 
-        for i in range(1,count+1):
-            try:
-                div:WebElement = self.driver.find_element_by_xpath(Paths.FOLLOWER_USER_DIV % i)
-                time.sleep(1)
-                username = div.text.split('\n')[0]
-                if  username not in followers:
-                    followers.append(username)
-                if i%callback_frequency==0:
-                    if callback is None:
-                        print('Got another {} followers...'.format(callback_frequency))
-                    else:
-                        callback(*args, **kwargs)
-                self.driver.execute_script("arguments[0].scrollIntoView();", div)
-                # TODO OPTIMIZE ALGORITHM (scroll by more than one account only)
-            except Exception as error:
-                raise error
-            
-        # and you're back
+        # Load all followers
+        followers = []
+        main:WebElement = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.FOLLOWERS_LIST_MAIN)))
+        size = main.size.get('height')
+        time.sleep(15)
+        while True:
+            main:WebElement = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.FOLLOWERS_LIST_MAIN)), wait_time=3)
+            new_size = main.size.get('height')
+            if new_size > 60000:
+                break
+            if new_size > size:
+                size = new_size
+                time.sleep(15)
+                continue
+            else:
+                break
+        followers_list:WebElement = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.FOLLOWERS_LIST)), wait_time=3)
+        divs = followers_list.find_elements_by_xpath(Paths.FOLLOWER_USER_DIV)
+        for div in divs:
+            username = div.text.split('\n')[0]
+            if username not in followers and username not in ('Follow',):
+                followers.append(username)
         return followers
-                
+
                 
     # IG UTILITY METHODS
     @insta_method
     def dismiss_dialogue(self):
         try:
-            dialogue = self._find_buttons(button_text='Not Now') # add this to 'Translation' doc
+            dialogue = self.__find_buttons(button_text='Not Now') # add this to 'Translation' doc
             dialogue.click()
         except:
             pass
@@ -362,7 +393,7 @@ class InstaClient:
         """
 
         self.driver.get(ClientUrls.SEARCH_TAGS.format(tag))
-        alert: WebElement = self._check_existence(EC.presence_of_element_located((By.XPATH, Paths.PAGE_NOT_FOUND)))
+        alert: WebElement = self.__check_existence(EC.presence_of_element_located((By.XPATH, Paths.PAGE_NOT_FOUND)))
         if alert:
             # Tag does not exist
             raise InvaildTagError(tag=tag)
@@ -372,12 +403,13 @@ class InstaClient:
 
 
     @insta_method
-    def nav_user(self, user:str):
+    def nav_user(self, user:str, check_user:bool=True):
         """
         Navigates to a users profile page
 
         Args:
             user:str: Username of the user to navigate to the profile page of
+            check_user:bool: Condition whether to check if a user is valid or not
 
         Returns:
             True if operation is successful
@@ -386,25 +418,13 @@ class InstaClient:
             InvaildUserError if user does not exist
         """
         self.driver.get(ClientUrls.NAV_USER.format(user))
-        element = self._check_existence(EC.presence_of_element_located((By.XPATH, Paths.PAGE_NOT_FOUND)), wait_time=3)
-        if element:
-            # User does not exist
-            self.driver.get(ClientUrls.HOME_URL)
-            print('Returned Home')
-            raise InvalidUserError(username=user)
-        else: 
-            # Operation Successful
-            print('Sucessfully Navigated to user')
-            paccount_alert = self._check_existence(EC.presence_of_element_located((By.XPATH, Paths.PRIVATE_ACCOUNT_ALERT)), wait_time=3)
-            if paccount_alert:
-                # navigate back to home page
-                raise PrivateAccountError(user)
-            else:
-                return True
+        if check_user:
+            return self.is_valid_user(user=user)
+        
 
 
     @insta_method
-    def nav_user_dm(self, user:str):
+    def nav_user_dm(self, user:str, check_user:bool=True):
         """
         Open DM page with a specific user
         
@@ -417,13 +437,31 @@ class InstaClient:
         Returns:
             True if operation was successful
         """
-        self.nav_user(user)
-        message_btn = self._find_buttons('Message')
+        self.nav_user(user, check_user=check_user)
+        message_btn = self.__find_buttons('Message')
         # Open User DM Page
         message_btn.click()
         return True
         
 
+    def is_valid_user(self, user):
+        element = self.__check_existence(EC.presence_of_element_located((By.XPATH, Paths.PAGE_NOT_FOUND)), wait_time=3)
+        if element:
+            # User does not exist
+            self.driver.get(ClientUrls.HOME_URL)
+            print('Returned Home')
+            raise InvalidUserError(username=user)
+        else: 
+            # Operation Successful
+            print('Sucessfully Navigated to user')
+            paccount_alert = self.__check_existence(EC.presence_of_element_located((By.XPATH, Paths.PRIVATE_ACCOUNT_ALERT)), wait_time=3)
+            if paccount_alert:
+                # navigate back to home page
+                raise PrivateAccountError(user)
+            else:
+                return True
+
+
     # IG PRIVATE UTILITIES
     def _infinite_scroll(self):
         """
@@ -452,18 +490,18 @@ class InstaClient:
         return False
 
 
-    def _find_buttons(self, button_text:str):
+    def __find_buttons(self, button_text:str):
         """
         Finds buttons for following and unfollowing users by filtering follow elements for buttons. Defaults to finding follow buttons.
 
         Args:
             button_text: Text that the desired button(s) has 
         """
-        buttons = self._find_element(EC.presence_of_element_located((By.XPATH, Paths.BUTTON.format(button_text))), wait_time=4)
+        buttons = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.BUTTON.format(button_text))), wait_time=4)
         return buttons
 
 
-    def _find_element(self, expectation, wait_time:int=10):
+    def __find_element(self, expectation, wait_time:int=10):
         """
         Finds widget (element) based on the field's value
         Args:
@@ -475,7 +513,7 @@ class InstaClient:
         return widgets
 
 
-    def _check_existence(self, expectation, wait_time:int=10):
+    def __check_existence(self, expectation, wait_time:int=10):
         """
         Checks if an element exists.
         Args:
